import { logger } from './logging.ts'
import { AppApiEvent, AppApiResponse } from './providers/jasd/appapi.dto.types.ts'

//const OPEN_API_BASE_URL = 'https://gemeinschaftswerk-nachhaltigkeit.de/app/openapi/v1'
const APP_API_BASE_URL = 'https://gemeinschaftswerk-nachhaltigkeit.de/app/api/v2'
const APP_FILES_BASE_URL = 'https://gemeinschaftswerk-nachhaltigkeit.de/app/api/v1/files'

export type HoliEventLocation = {
  street?: string
  streetNo?: string
  zipCode?: string
  city?: string
}

export type HoliEventDetails = {
  description: string
  externalLink?: string
  meetingLink?: string
}

export type HoliEvent = {
  id: string
  title: string
  organisationName: string
  startDate: string | null
  endDate: string | null
  location: HoliEventLocation
  isRemote: boolean
  imageUrl: string
  eventDetails: HoliEventDetails
}

export type HoliEventsResponse = {
  totalResults: number
  data: HoliEvent[]
}

const toHoliEvent = (event: AppApiEvent): HoliEvent => {
  return {
    id: event.activity.id.toString(),
    title: event.name,
    organisationName: event.activity.organisation.name,
    startDate: event.activity.period.start,
    endDate: event.activity.period.end,
    location: {
      city: event.location.address.city,
      street: event.location.address.street,
      streetNo: event.location.address.streetNo,
      zipCode: event.location.address.zipCode,
    },
    isRemote: event.activity.location.online,
    // todo gregor: imageproxy-fi the url
    imageUrl: APP_FILES_BASE_URL + '/' + event.activity.image,
    eventDetails: {
      description: event.description,
      externalLink: event.activity.location.url,
      meetingLink: event.activity.registerUrl,
    },
  }
}

const buildJasdAppApiUrl = (
  { limit, offset, location, remoteOnly }: FetchEventsInput,
): URL => {
  const url = new URL(`${APP_API_BASE_URL}/search`)
  url.searchParams.append('page', offset.toString())
  url.searchParams.append('size', limit.toString())
  url.searchParams.append('sort', 'period.start,asc,ignorecase')
  // filtering outdated and permanent events here helps us a lot
  // and provides only OnlineTimedEvent and RealTimedEvent which have start and end dates
  url.searchParams.append('includeExpiredActivities', 'false')
  url.searchParams.append('permanent', 'false')

  if (remoteOnly) {
    url.searchParams.append('online', 'true')
  }

  if (location) {
    url.searchParams.append('location', location)
  }

  return url
}

type EventsPage = {
  events: HoliEvent[]
  onLastPage: boolean
  totalResults: number
}

const fetchEventsPage = async (
  url: URL,
): Promise<EventsPage> => {
  const startDate = Date.now()

  logger.debug(`fetching events from ${url}`)
  try {
    const response = await fetch(url, {
      // gregor: we don't need to send the api key to the App api
      // only the OpenApi endpoints require auth
      // headers: { 'x-api-key': Deno.env.get('JASD_DEVELOPER_API_KEY')! },
    })
    const json = await response.json() as AppApiResponse

    const holiEvents = json.content
      .flat()
      .filter((event) => !event.activity.period.permanent)
      .map(toHoliEvent)
      .filter((holiEvent) => !!holiEvent)

    return {
      events: holiEvents,
      onLastPage: json.last,
      totalResults: json.totalElements,
    }
    // deno-lint-ignore no-explicit-any
  } catch (e: any) {
    logger.error(
      `Error performing request to ${url}: ${e.message}`,
    )
    throw e
  } finally {
    const duration = Date.now() - startDate
    logger.info(`fetching projects took ${duration} ms`)
  }
}

export type FetchEventsInput = {
  limit: number
  offset: number
  location?: string
  remoteOnly?: boolean
}

export const fetchEvents = async (
  input: FetchEventsInput,
): Promise<HoliEventsResponse> => {
  const url = buildJasdAppApiUrl(input)
  const fetchResult = await fetchEventsPage(url)

  return {
    totalResults: fetchResult.totalResults,
    data: fetchResult.events,
  }
}
