import { logger } from './logging.ts'
import { fetchEvents, FetchEventsInput, HoliEventsResponse } from './fetch_events.ts'
import { fetchEvent, FetchEventInput, HoliEventResponse } from './fetch_event.ts'
import { GeoApiClient } from './providers/geo/GeoApiClient.ts'
import { createSchema, createYoga } from './deps.ts'

const SCHEMA = `
    type Location {
      street: String
      streetNo: String
      zipCode: String
      city: String
    }
    
    type EventDetails {
      description: String!
      externalLink: String
      meetingLink: String
    }
  
    type Event {
      id: String!
      title: String!
      organisationName: String!
      startDate: String
      endDate: String
      location: Location
      isRemote: Boolean!
      imageUrl: String!
      eventDetails: EventDetails!
    }
    
    type EventsResponse {
      totalResults: Int!
      data: [Event]
    }

    type EventResponse {
      data: Event
    }
    
    type Query {
      # uses offset-based pagination as described in https://www.apollographql.com/docs/react/pagination/offset-based
      events(offset: Int = 0, limit: Int = 5, location: String, remoteOnly: Boolean): EventsResponse!
      event(id: String!): EventResponse!
    }
`

export type EventsRequest = {
  limit?: number
  offset?: number
  location?: string
  remoteOnly?: boolean
}

export type EventRequest = {
  id: string
}

const validateEventsRequestInput = (request: EventsRequest): FetchEventsInput => {
  logger.debug('validating request: ' + JSON.stringify(request))

  let limit = request.limit ?? 5
  if (limit > 50) {
    limit = 50
  }

  return {
    limit: limit,
    offset: request.offset ?? 0,
    location: request.location,
    remoteOnly: request.remoteOnly,
  }
}

const validateEventRequestInput = (request: EventRequest): FetchEventInput => {
  logger.debug('validating request: ' + JSON.stringify(request))
  return {
    id: request.id,
  }
}

const createResolvers = (config: ServerConfig) => ({
  Query: {
    events: (
      // deno-lint-ignore no-explicit-any
      _parent: any,
      parameters: EventsRequest,
      // next line is required for the resolver to work
      // deno-lint-ignore no-unused-vars
      context: GraphQLContext,
    ): Promise<HoliEventsResponse> => {
      if (config.fake) {
        return Promise.resolve({ totalResults: 0, data: [] })
      } else {
        return fetchEvents(
          validateEventsRequestInput(parameters),
        )
      }
    },
    event: (
      // deno-lint-ignore no-explicit-any
      _parent: any,
      parameters: EventRequest,
      // next line is required for the resolver to work
      // deno-lint-ignore no-unused-vars
      context: GraphQLContext,
    ): Promise<HoliEventResponse> => {
      if (config.fake) {
        return Promise.resolve({ data: {} })
      } else {
        return fetchEvent(
          validateEventRequestInput(parameters),
        )
      }
    },
  },
})

export const DEFAULT_PORT = 8006

export type ServerConfig = {
  port: number
  fake: boolean // For local development. If set, the API returns dummy data
  imageProxyBaseUrl: string
  geoAPIEndpointUrl: string
}

const isAlive = () => new Response(`${true}`)

type GraphQLContext = {
  request?: { headers: Headers }
  params?: {
    extensions?: {
      headers?: {
        [key: string]: string
      }
    }
  }
}

export const startServer = (config: ServerConfig): Deno.HttpServer<Deno.NetAddr> => {
  const _ = new GeoApiClient(config.geoAPIEndpointUrl)

  const resolvers = createResolvers(config)

  const gqlServer = createYoga({
    schema: createSchema({ resolvers, typeDefs: SCHEMA }),
    graphiql: true,
  })

  return Deno.serve({
    port: config.port,
    handler: (req: Request) => {
      const url = new URL(req.url)
      const pathname = url.pathname
      if (pathname.startsWith('/health') || pathname.startsWith('/ready')) return isAlive()
      return gqlServer.handle(req)
    },
    onListen({ port, hostname }) {
      logger.info(
        `Server started at http://${hostname === '0.0.0.0' ? 'localhost' : hostname}:${port}/graphql`,
      )
      if (config.fake) {
        logger.info(
          'Server is serving fake data due to FAKE env var set to true',
        )
      }
    },
  })
}
